use crate::todo::{ChangeType, TaskColumn, TaskListDB};
use rusqlite::{params, Connection, Error, Result};
use std::path::Path;

const DEFAULT_NAME: &str = "general";

pub(crate) fn startup(path: &Path) -> std::result::Result<Connection, Error> {
    let conn = Connection::open(path)?;

    conn.execute(
        "CREATE TABLE IF NOT EXISTS task_group (
                id INTEGER PRIMARY KEY,
                value TEXT NOT NULL,
                UNIQUE(value)
        )",
        ()
    )?;

    conn.execute(
        "INSERT OR IGNORE INTO task_group (value) VALUES (?1)",
        (&DEFAULT_NAME,)
    )?;

    if let Err(e) = create_tasklist_tables(&conn) {
        eprintln!("[ERROR] Failed to create tasklist tables:\n{e}");
        return Err(e);
    }
    
    Ok(conn)
}

fn create_tasklist_tables(conn: &Connection) -> Result<()> {
    let tl = get_tasklists(&conn)?;
    if tl.len() < 1 {
        return Ok(());
    }
    
    for item in 1..tl.len()+1 {
        let stmt = format!("CREATE TABLE IF NOT EXISTS t{item} (
                id INTEGER PRIMARY KEY,
                checked INTEGER NOT NULL,
                value TEXT NOT NULL
        )");
        
        conn.execute(
            &stmt.to_string(),
            ()
        )?;
    }
    
    Ok(())
}

pub(crate) fn get_tasks(conn: &Connection, tl_id: usize) -> std::result::Result<Vec<TaskColumn>, Error> {
    let stmt_format = format!("SELECT id, checked, value FROM t{tl_id}");
    let mut stmt = conn.prepare(stmt_format.as_ref())?;
    let tasks_iter = stmt.query_map([], |row| {
        Ok(TaskColumn {
            id: row.get(0)?,
            checked: row.get(1)?,
            value: row.get(2)?,
        })
    })?;
    
    let mut tasks = Vec::new();
    for task in tasks_iter {
        tasks.push(task?);
    }

    Ok(tasks)
}

pub(crate) fn insert_task(conn: &Connection, task: TaskColumn) -> Result<i64> {
    let stmt = format!("INSERT INTO t{} (checked, value) VALUES (?1, ?2)", &task.id);
    
    conn.execute(
        &stmt,
        (&task.checked, &task.value),
    )?;
    
    Ok(conn.last_insert_rowid())
}

pub(crate) fn update_task(conn: &Connection, task: TaskColumn, change: ChangeType, tl_id: usize) -> Result<()> {
    let val = format!("UPDATE t{tl_id} SET value = ?1 WHERE id = ?2");
    let check = format!("UPDATE t{tl_id} SET checked = ?1 WHERE id = ?2");
    
    let (query, params) = match change {
        ChangeType::Checked => (
            check.as_ref(),
            params![&task.checked, &task.id]
        ),
        ChangeType::Value => (
            val.as_ref(),
            params![&task.value, &task.id]
        ),
    };

    conn.execute(query, params)?;
    Ok(())
}

pub(crate) fn update_unselect_tasks(conn: &Connection, checked: bool, tl_id: usize) -> Result<()> {
    let stmt = format!("UPDATE t{tl_id} SET checked = ?1");
    
    conn.execute(
        stmt.as_ref(),
        (&checked,)
    )?;
    
    Ok(())
}

pub(crate) fn delete_tasks(conn: &Connection, id: Option<usize>, tl_id: usize) -> Result<()> {
    let del_id = format!("DELETE FROM t{tl_id} WHERE id = ?1").to_string();
    let full_del = format!("DELETE FROM t{tl_id}").to_string();
    
    let (query, params) = match id { 
        Some(t) => (
            del_id.as_ref(),
            params![t.clone()]
        ),
        None => (
            full_del.as_ref(),
            params![]
        ),
    };

    conn.execute(query, params)?;
    Ok(())
}

pub(crate) fn insert_tasklist(conn: &Connection, task_name: &str) -> Result<i64> {
    conn.execute(
        "INSERT OR IGNORE INTO task_group (value) VALUES (?1)",
        (&task_name,)
    )?;
    
    Ok(conn.last_insert_rowid())
}

pub(super) fn get_tasklists(conn: &Connection) -> std::result::Result<Vec<TaskListDB>, Error> {
    let mut stmt = conn.prepare("SELECT id, value FROM task_group")?;
    let tasks_iter = stmt.query_map([], |row| {
        Ok(TaskListDB {
            id: row.get(0)?,
            value: row.get(1)?,
        })
    })?;

    let mut tasks = Vec::new();
    for task in tasks_iter {
        tasks.push(task?);
    }

    Ok(tasks)
}